//Code by Robert Pohl Nov. 2003
//robert@2cool2care.com
//Cred to: called odah and Parvez Ahmad Hakim.
//They have written the code about getting icons and drive information with the Interop service. 

using System;
using System.Text;
using System.IO;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Runtime.InteropServices;


namespace DirectoryPicker
{
	public enum DirectoryPickerResponse
	{
		Ok,Cancel
	}
	[StructLayout(LayoutKind.Sequential)]
	public struct SHFILEINFO 
	{
		public IntPtr hIcon;
		public IntPtr iIcon;
		public uint dwAttributes;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
		public string szDisplayName;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 80)]
		public string szTypeName;
	};

	class Win32
	{
		public const uint SHGFI_ICON = 0x100;
		public const uint SHGFI_LARGEICON = 0x0;    // 'Large icon
		public const uint SHGFI_SMALLICON = 0x1;    // 'Small icon

		[DllImport("shell32.dll")]
		public static extern IntPtr SHGetFileInfo(string pszPath,
			uint dwFileAttributes,
			ref SHFILEINFO psfi,
			uint cbSizeFileInfo,
			uint uFlags);
	}

	public class DirPicker : System.Windows.Forms.Form
	{
		private System.Windows.Forms.TreeView treeView1;
		private ImageList myImageList = new ImageList();
		private bool isLocked = false;
		private string Title = "Select Directory";

		private Button OkButton = new Button();
		private Button CnclButton = new Button();
		public DirectoryPickerResponse Result = new DirectoryPickerResponse();
		public string SelectedDirectory = "";

		[DllImport("kernel32.dll")]
		public static extern long GetDriveType(string driveLetter);
		[DllImport("kernel32.dll")]
		public static extern long GetVolumeInformation(string strPathName,
			StringBuilder strVolumeNameBuffer,
			long lngVolumeNameSize,
			long lngVolumeSerialNumber,
			long lngMaximumComponentLength,
			long lngFileSystemFlags,
			StringBuilder strFileSystemNameBuffer,
			long lngFileSystemNameSize);

		private System.ComponentModel.Container components = null;

		public DirPicker()
		{
			InitializeComponent();

			CreateFolderIcon();
			this.Text = this.Title;
			this.Result = DirectoryPickerResponse.Cancel;

			this.Resize +=new EventHandler(Form1_Resize);
			CnclButton.Text = "Cancel";
			CnclButton.Top = this.Height-CnclButton.Height-40;
			CnclButton.Left = 15;
			CnclButton.Click +=new EventHandler(CnclButton_Click);
			this.Controls.Add(CnclButton);

			OkButton.Text = "Select";
			OkButton.Top = this.Height-OkButton.Height-40;
			OkButton.Left = 195;
			OkButton.Click+=new EventHandler(OkButton_Click);
			this.Controls.Add(OkButton);

			Label HeadLine = new Label();
			HeadLine.Text = "Select Directory";
			HeadLine.Top = 15;
			HeadLine.Left = 15;
			this.Controls.Add(HeadLine);

			//when treeview is clicked
			this.treeView1.MouseUp+=new MouseEventHandler(treeView1_Click);
			this.treeView1.BeforeExpand +=new TreeViewCancelEventHandler(treeView1_BeforeExpand);
			this.treeView1.BeforeCollapse+=new TreeViewCancelEventHandler(treeView1_BeforeCollapse);
			myImageList.ColorDepth = ColorDepth.Depth16Bit;
			myImageList.TransparentColor = Color.Transparent;
			
			//Generate list of system drives
			string[] drives = Directory.GetLogicalDrives();
			string _name ="";
			TreeNode tmpNode;

			//Loop through system drives and get name and path.
			//Name is the TreeNode's Text and the path is in the Tag property.
			foreach (string s in drives)
			{
				_name = GetDriveName(s);
				if(getDriveType(s)==4) _name="Network Drive";
				if(getDriveType(s)==5) _name="CD-ROM/DVD";

				myImageList.Images.Add(GetIcon(s));
				tmpNode = new TreeNode(_name+" ("+s+")",myImageList.Images.Count-1,myImageList.Images.Count-1);
				tmpNode.Tag = s;
				//Since A: and CD-ROM is not always readable we have to try n catch it.
				if(this.getDriveType(s)!=5&&this.getDriveType(s)!=2)
				{
					try
					{
						HasSubDirs(tmpNode);
					}
					catch{}
				}
				this.treeView1.Nodes.Add(tmpNode);
			}

			this.treeView1.ImageList = myImageList;
		}
		#region Event Handlers. If we expand/collapse, the node method is locked 
		//If we press cancel button set the result to cancel, and close.
		private void CnclButton_Click(object sender, EventArgs e)
		{
			this.Result = DirectoryPickerResponse.Cancel;
			this.Close();
		}
		//If we select a directory set the result to ok, set the selected dir and close
		private void OkButton_Click(object sender, EventArgs e)
		{
			this.Result = DirectoryPickerResponse.Ok;
			this.SelectedDirectory = this.treeView1.SelectedNode.Tag.ToString();
			this.Close();
		}
		//Since the expand/collapse function interfere with the "treeView1_Click" handler, we need to 
		//make this lock. Mabe you ca do it better? :)
		private void treeView1_BeforeExpand(object sender, System.Windows.Forms.TreeViewCancelEventArgs e)
		{
			this.isLocked = true;
			ViewFolder(e.Node);
		}
		private void treeView1_BeforeCollapse(object sender, System.Windows.Forms.TreeViewCancelEventArgs e)
		{
			this.isLocked = true;
		}
		public void treeView1_Click(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			//In case that we have collapsed/expanded a node just before, we don't want to trigger this method.
			if(this.isLocked==false)
			{
				/* The clicked node */
				if(IsNode(this.treeView1.GetNodeAt(this.treeView1.PointToClient(Cursor.Position)))==true)
				{
					ClickNode(this.treeView1.GetNodeAt(this.treeView1.PointToClient(Cursor.Position)));
				}
			}
			else //First unlock the method.
			{
				this.isLocked = false;
			}
		}
		private void Form1_Resize(object sender, EventArgs e)
		{
			this.treeView1.Width = this.Width-50;
			this.treeView1.Height = this.Height-this.treeView1.Top-100;
			this.OkButton.Top = this.treeView1.Height+this.treeView1.Top+25;
			this.OkButton.Left = this.treeView1.Width+this.treeView1.Left-this.OkButton.Width;
			this.CnclButton.Top = this.treeView1.Height+this.treeView1.Top+25;
		}

		#endregion

		//Looks for the first dir in the first HD, and takes that icon
		private void CreateFolderIcon()
		{
				string[] drives = Directory.GetLogicalDrives();
			int i = 0;
			for(i=0;i<drives.Length;i++)
			{
				//look for the first harddrive
				if(getDriveType(drives[i])==3)
				{
					break;
				}
			}
			string[] dir = Directory.GetDirectories(drives[i]);
			this.myImageList.Images.Add(GetIcon(dir[0]));
		}
		//This method returns a ICON from a path
		private Icon GetIcon(string _file)
		{
			SHFILEINFO shinfo = new SHFILEINFO();
			IntPtr hImgSmall;    //the handle to the system image list
			hImgSmall = Win32.SHGetFileInfo(_file, 0, ref shinfo,
				(uint)Marshal.SizeOf(shinfo),
				Win32.SHGFI_ICON |
				Win32.SHGFI_SMALLICON);
			return System.Drawing.Icon.FromHandle(shinfo.hIcon);
		}
		
		//Generates childnodes, if possibe
		private void ClickNode(TreeNode _node)
		{
			this.treeView1.SelectedNode = _node;
			try
			{
				ViewFolder(_node);
			}
			catch (Exception e)
			{
				MessageBox.Show(e.Message);
			}
			this.treeView1.SelectedNode.Expand();
		}
		private void ViewFolder(TreeNode _node)
		{
			_node.Nodes.Clear();

			TreeNode tmpNode;
			string[] dir = Directory.GetDirectories(_node.Tag.ToString());

			foreach(string d in dir)
			{
				tmpNode = new TreeNode(d.Substring(d.LastIndexOf(@"\")+1));
				tmpNode.Tag = d;
				//This (HasSubDirs) makes about the same as the method we are already in,
				//but we don't want to loop through ALL folders.. that will take to long :)
				try
				{
					HasSubDirs(tmpNode);
				}
				catch{}

				_node.Nodes.Add(tmpNode);
			}
		}

		private void HasSubDirs(TreeNode _node)
		{
			_node.Nodes.Clear();

			TreeNode tmpNode;
			string[] dir = Directory.GetDirectories(_node.Tag.ToString());

			for(int i=0;i<dir.Length;i++)
			{
				tmpNode = new TreeNode(dir[i].ToString());
				tmpNode.Tag = dir[i].ToString();
				_node.Nodes.Add(tmpNode);
			}
		}

		public bool IsNode(TreeNode mynode)
		{
			//Are we clickin on a node?
			try
			{
				string foo = mynode.Text;
				return true;
			}
			catch
			{
				return false;
			}
		}

		public string GetDriveName(string drive)
		{
			//receives volume name of drive
			StringBuilder volname = new StringBuilder(256);
			//receives serial number of drive,not in case of network drive(win95/98)
			long sn= new long();
			long maxcomplen = new long();//receives maximum component length
			long sysflags = new long();//receives file system flags
			StringBuilder sysname = new StringBuilder(256);//receives the file system name
			long retval= new long();//return value

			retval = GetVolumeInformation(drive, volname, 256, sn, maxcomplen, 
				sysflags, sysname,256);
            
			if(retval!=0)return volname.ToString();
			else return "";
		}
		public int getDriveType(string drive)
		{
			if((GetDriveType(drive) & 5)==5)return 5;//cd
			if((GetDriveType(drive) & 3)==3)return 3;//fixed
			if((GetDriveType(drive) & 2)==2)return 2;//removable
			if((GetDriveType(drive) & 4)==4)return 4;//remote disk
			if((GetDriveType(drive) & 6)==6)return 6;//ram disk
			return 0;
		}


		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.Size = new System.Drawing.Size(300,300);

			this.treeView1 = new System.Windows.Forms.TreeView();
			this.treeView1.ImageIndex = -1;
			this.treeView1.Location = new System.Drawing.Point(16, 56);
			this.treeView1.Name = "treeView1";
			this.treeView1.SelectedImageIndex = -1;
			this.treeView1.Size = new System.Drawing.Size(256, 160);
			this.treeView1.TabIndex = 0;
			this.Controls.Add(this.treeView1);


		}
		#endregion
	}
}
